"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createFirebaseEndpoints = void 0;
const emulatorLogger_1 = require("../../emulatorLogger");
const types_1 = require("../../types");
const zlib_1 = require("zlib");
const metadata_1 = require("../metadata");
const express_1 = require("express");
const registry_1 = require("../../registry");
const multipart_1 = require("../multipart");
const errors_1 = require("../errors");
const upload_1 = require("../upload");
function createFirebaseEndpoints(emulator) {
    const firebaseStorageAPI = (0, express_1.Router)();
    const { storageLayer, uploadService } = emulator;
    if (process.env.STORAGE_EMULATOR_DEBUG) {
        firebaseStorageAPI.use((req, res, next) => {
            console.log("--------------INCOMING REQUEST--------------");
            console.log(`${req.method.toUpperCase()} ${req.path}`);
            console.log("-- query:");
            console.log(JSON.stringify(req.query, undefined, 2));
            console.log("-- headers:");
            console.log(JSON.stringify(req.headers, undefined, 2));
            console.log("-- body:");
            if (req.body instanceof Buffer) {
                console.log(`Buffer of ${req.body.length}`);
            }
            else if (req.body) {
                console.log(req.body);
            }
            else {
                console.log("Empty body (could be stream)");
            }
            const resJson = res.json.bind(res);
            res.json = (...args) => {
                console.log("-- response:");
                args.forEach((data) => console.log(JSON.stringify(data, undefined, 2)));
                return resJson.call(res, ...args);
            };
            const resSendStatus = res.sendStatus.bind(res);
            res.sendStatus = (status) => {
                console.log("-- response status:");
                console.log(status);
                return resSendStatus.call(res, status);
            };
            const resStatus = res.status.bind(res);
            res.status = (status) => {
                console.log("-- response status:");
                console.log(status);
                return resStatus.call(res, status);
            };
            next();
        });
    }
    firebaseStorageAPI.use(/.*\/b\/(.+?)\/.*/, (req, res, next) => {
        const bucketId = req.params[0];
        storageLayer.createBucket(bucketId);
        if (!emulator.rulesManager.getRuleset(bucketId)) {
            emulatorLogger_1.EmulatorLogger.forEmulator(types_1.Emulators.STORAGE).log("WARN", "Permission denied because no Storage ruleset is currently loaded, check your rules for syntax errors.");
            return res.status(403).json({
                error: {
                    code: 403,
                    message: "Permission denied. Storage Emulator has no loaded ruleset.",
                },
            });
        }
        next();
    });
    firebaseStorageAPI.get("/b/:bucketId/o/:objectId", async (req, res) => {
        var _a;
        let metadata;
        let data;
        try {
            ({ metadata, data } = await storageLayer.handleGetObject({
                bucketId: req.params.bucketId,
                decodedObjectId: decodeURIComponent(req.params.objectId),
                authorization: req.header("authorization"),
                downloadToken: (_a = req.query.token) === null || _a === void 0 ? void 0 : _a.toString(),
            }));
        }
        catch (err) {
            if (err instanceof errors_1.NotFoundError) {
                return res.sendStatus(404);
            }
            else if (err instanceof errors_1.ForbiddenError) {
                return res.status(403).json({
                    error: {
                        code: 403,
                        message: `Permission denied. No READ permission.`,
                    },
                });
            }
            throw err;
        }
        if (!metadata.downloadTokens.length) {
            metadata.addDownloadToken();
        }
        if (req.query.alt === "media") {
            const isGZipped = metadata.contentEncoding === "gzip";
            if (isGZipped) {
                data = (0, zlib_1.gunzipSync)(data);
            }
            res.setHeader("Accept-Ranges", "bytes");
            res.setHeader("Content-Type", metadata.contentType);
            setObjectHeaders(res, metadata, { "Content-Encoding": isGZipped ? "identity" : undefined });
            const byteRange = req.range(data.byteLength, { combine: true });
            if (Array.isArray(byteRange) && byteRange.type === "bytes" && byteRange.length > 0) {
                const range = byteRange[0];
                res.setHeader("Content-Range", `${byteRange.type} ${range.start}-${range.end}/${data.byteLength}`);
                res.status(206).end(data.slice(range.start, range.end + 1));
            }
            else {
                res.end(data);
            }
            return;
        }
        return res.json(new metadata_1.OutgoingFirebaseMetadata(metadata));
    });
    firebaseStorageAPI.get("/b/:bucketId/o", async (req, res) => {
        var _a, _b;
        const maxResults = (_a = req.query.maxResults) === null || _a === void 0 ? void 0 : _a.toString();
        let response;
        try {
            response = await storageLayer.handleListObjects({
                bucketId: req.params.bucketId,
                prefix: req.query.prefix ? req.query.prefix.toString() : "",
                delimiter: req.query.delimiter ? req.query.delimiter.toString() : "",
                pageToken: (_b = req.query.pageToken) === null || _b === void 0 ? void 0 : _b.toString(),
                maxResults: maxResults ? +maxResults : undefined,
                authorization: req.header("authorization"),
            });
        }
        catch (err) {
            if (err instanceof errors_1.ForbiddenError) {
                return res.status(403).json({
                    error: {
                        code: 403,
                        message: `Permission denied. No LIST permission.`,
                    },
                });
            }
            throw err;
        }
        return res.json(response);
    });
    const reqBodyToBuffer = async (req) => {
        if (req.body instanceof Buffer) {
            return Buffer.from(req.body);
        }
        const bufs = [];
        req.on("data", (data) => {
            bufs.push(data);
        });
        await new Promise((resolve) => {
            req.on("end", () => {
                resolve();
            });
        });
        return Buffer.concat(bufs);
    };
    const handleUpload = async (req, res) => {
        if (!req.query.name) {
            res.sendStatus(400);
            return;
        }
        const bucketId = req.params.bucketId;
        const objectId = req.query.name.toString();
        const uploadType = req.header("x-goog-upload-protocol");
        if (uploadType === "multipart") {
            const contentTypeHeader = req.header("content-type");
            if (!contentTypeHeader) {
                return res.sendStatus(400);
            }
            let metadataRaw;
            let dataRaw;
            try {
                ({ metadataRaw, dataRaw } = (0, multipart_1.parseObjectUploadMultipartRequest)(contentTypeHeader, await reqBodyToBuffer(req)));
            }
            catch (err) {
                if (err instanceof Error) {
                    return res.status(400).json({
                        error: {
                            code: 400,
                            message: err.toString(),
                        },
                    });
                }
                throw err;
            }
            const upload = uploadService.multipartUpload({
                bucketId,
                objectId,
                metadataRaw,
                dataRaw: dataRaw,
                authorization: req.header("authorization"),
            });
            let metadata;
            try {
                metadata = await storageLayer.handleUploadObject(upload);
            }
            catch (err) {
                if (err instanceof errors_1.ForbiddenError) {
                    return res.status(403).json({
                        error: {
                            code: 403,
                            message: "Permission denied. No WRITE permission.",
                        },
                    });
                }
                throw err;
            }
            metadata.addDownloadToken();
            return res.status(200).json(new metadata_1.OutgoingFirebaseMetadata(metadata));
        }
        const uploadCommand = req.header("x-goog-upload-command");
        if (!uploadCommand) {
            res.sendStatus(400);
            return;
        }
        if (uploadCommand === "start") {
            const upload = uploadService.startResumableUpload({
                bucketId,
                objectId,
                metadataRaw: JSON.stringify(req.body),
                authorization: req.header("authorization"),
            });
            res.header("x-goog-upload-chunk-granularity", "10000");
            res.header("x-goog-upload-control-url", "");
            res.header("x-goog-upload-status", "active");
            const emulatorInfo = registry_1.EmulatorRegistry.getInfo(types_1.Emulators.STORAGE);
            res.header("x-goog-upload-url", `http://${req.hostname}:${emulatorInfo === null || emulatorInfo === void 0 ? void 0 : emulatorInfo.port}/v0/b/${bucketId}/o?name=${objectId}&upload_id=${upload.id}&upload_protocol=resumable`);
            res.header("x-gupload-uploadid", upload.id);
            return res.sendStatus(200);
        }
        if (!req.query.upload_id) {
            return res.sendStatus(400);
        }
        const uploadId = req.query.upload_id.toString();
        if (uploadCommand === "query") {
            let upload;
            try {
                upload = uploadService.getResumableUpload(uploadId);
            }
            catch (err) {
                if (err instanceof errors_1.NotFoundError) {
                    return res.sendStatus(404);
                }
                throw err;
            }
            res.header("X-Goog-Upload-Size-Received", upload.size.toString());
            return res.sendStatus(200);
        }
        if (uploadCommand === "cancel") {
            try {
                uploadService.cancelResumableUpload(uploadId);
            }
            catch (err) {
                if (err instanceof errors_1.NotFoundError) {
                    return res.sendStatus(404);
                }
                else if (err instanceof upload_1.NotCancellableError) {
                    return res.sendStatus(400);
                }
                throw err;
            }
            return res.sendStatus(200);
        }
        if (uploadCommand.includes("upload")) {
            let upload;
            try {
                upload = uploadService.continueResumableUpload(uploadId, await reqBodyToBuffer(req));
            }
            catch (err) {
                if (err instanceof errors_1.NotFoundError) {
                    return res.sendStatus(404);
                }
                else if (err instanceof upload_1.UploadNotActiveError) {
                    return res.sendStatus(400);
                }
                throw err;
            }
            if (!uploadCommand.includes("finalize")) {
                res.header("x-goog-upload-status", "active");
                res.header("x-gupload-uploadid", upload.id);
                return res.sendStatus(200);
            }
        }
        if (uploadCommand.includes("finalize")) {
            let upload;
            try {
                upload = uploadService.finalizeResumableUpload(uploadId);
            }
            catch (err) {
                if (err instanceof errors_1.NotFoundError) {
                    return res.sendStatus(404);
                }
                else if (err instanceof upload_1.UploadNotActiveError) {
                    return res.sendStatus(400);
                }
                throw err;
            }
            let metadata;
            try {
                metadata = await storageLayer.handleUploadObject(upload);
            }
            catch (err) {
                if (err instanceof errors_1.ForbiddenError) {
                    return res.status(403).json({
                        error: {
                            code: 403,
                            message: `Permission denied. No WRITE permission.`,
                        },
                    });
                }
                throw err;
            }
            metadata.addDownloadToken();
            return res.status(200).json(new metadata_1.OutgoingFirebaseMetadata(metadata));
        }
        return res.sendStatus(400);
    };
    const handleTokenRequest = (req, res) => {
        var _a, _b;
        if (!req.query.create_token && !req.query.delete_token) {
            return res.sendStatus(400);
        }
        const bucketId = req.params.bucketId;
        const decodedObjectId = decodeURIComponent(req.params.objectId);
        const authorization = req.header("authorization");
        let metadata;
        if (req.query.create_token) {
            if (req.query.create_token !== "true") {
                return res.sendStatus(400);
            }
            try {
                metadata = storageLayer.handleCreateDownloadToken({
                    bucketId,
                    decodedObjectId,
                    authorization,
                });
            }
            catch (err) {
                if (err instanceof errors_1.ForbiddenError) {
                    return res.status(403).json({
                        error: {
                            code: 403,
                            message: `Missing admin credentials.`,
                        },
                    });
                }
                if (err instanceof errors_1.NotFoundError) {
                    return res.sendStatus(404);
                }
                throw err;
            }
        }
        else {
            try {
                metadata = storageLayer.handleDeleteDownloadToken({
                    bucketId,
                    decodedObjectId,
                    token: (_b = (_a = req.query["delete_token"]) === null || _a === void 0 ? void 0 : _a.toString()) !== null && _b !== void 0 ? _b : "",
                    authorization,
                });
            }
            catch (err) {
                if (err instanceof errors_1.ForbiddenError) {
                    return res.status(403).json({
                        error: {
                            code: 403,
                            message: `Missing admin credentials.`,
                        },
                    });
                }
                if (err instanceof errors_1.NotFoundError) {
                    return res.sendStatus(404);
                }
                throw err;
            }
        }
        setObjectHeaders(res, metadata);
        return res.json(new metadata_1.OutgoingFirebaseMetadata(metadata));
    };
    const handleObjectPostRequest = async (req, res) => {
        if (req.query.create_token || req.query.delete_token) {
            return handleTokenRequest(req, res);
        }
        return handleUpload(req, res);
    };
    const handleMetadataUpdate = async (req, res) => {
        let metadata;
        try {
            metadata = await storageLayer.handleUpdateObjectMetadata({
                bucketId: req.params.bucketId,
                decodedObjectId: decodeURIComponent(req.params.objectId),
                metadata: req.body,
                authorization: req.header("authorization"),
            });
        }
        catch (err) {
            if (err instanceof errors_1.ForbiddenError) {
                return res.status(403).json({
                    error: {
                        code: 403,
                        message: `Permission denied. No WRITE permission.`,
                    },
                });
            }
            if (err instanceof errors_1.NotFoundError) {
                return res.sendStatus(404);
            }
            throw err;
        }
        setObjectHeaders(res, metadata);
        return res.json(new metadata_1.OutgoingFirebaseMetadata(metadata));
    };
    firebaseStorageAPI.patch("/b/:bucketId/o/:objectId", handleMetadataUpdate);
    firebaseStorageAPI.put("/b/:bucketId/o/:objectId?", async (req, res) => {
        var _a;
        switch ((_a = req.header("x-http-method-override")) === null || _a === void 0 ? void 0 : _a.toLowerCase()) {
            case "patch":
                return handleMetadataUpdate(req, res);
            default:
                return handleObjectPostRequest(req, res);
        }
    });
    firebaseStorageAPI.post("/b/:bucketId/o/:objectId?", handleObjectPostRequest);
    firebaseStorageAPI.delete("/b/:bucketId/o/:objectId", async (req, res) => {
        try {
            await storageLayer.handleDeleteObject({
                bucketId: req.params.bucketId,
                decodedObjectId: decodeURIComponent(req.params.objectId),
                authorization: req.header("authorization"),
            });
        }
        catch (err) {
            if (err instanceof errors_1.ForbiddenError) {
                return res.status(403).json({
                    error: {
                        code: 403,
                        message: `Permission denied. No WRITE permission.`,
                    },
                });
            }
            if (err instanceof errors_1.NotFoundError) {
                return res.sendStatus(404);
            }
            throw err;
        }
        res.sendStatus(204);
    });
    firebaseStorageAPI.get("/", (req, res) => {
        res.json({ emulator: "storage" });
    });
    return firebaseStorageAPI;
}
exports.createFirebaseEndpoints = createFirebaseEndpoints;
function setObjectHeaders(res, metadata, headerOverride = { "Content-Encoding": undefined }) {
    res.setHeader("Content-Disposition", metadata.contentDisposition);
    if (headerOverride["Content-Encoding"]) {
        res.setHeader("Content-Encoding", headerOverride["Content-Encoding"]);
    }
    else {
        res.setHeader("Content-Encoding", metadata.contentEncoding);
    }
    if (metadata.cacheControl) {
        res.setHeader("Cache-Control", metadata.cacheControl);
    }
    if (metadata.contentLanguage) {
        res.setHeader("Content-Language", metadata.contentLanguage);
    }
}
